"""
Apply normalization to given input
"""

import sys
from . import NormalizationComposite
import argparse
from . import factory
from .logger import DiffLoggingFormatter, Logger
import logging
from benchmarkstt.cli import args_from_factory
from benchmarkstt import settings


def args_inputfile(parser):
    parser.add_argument('-i', '--inputfile', action='append', nargs=1,
                        help='read input from this file, defaults to STDIN',
                        metavar='file')


def args_logs(parser: argparse.ArgumentParser):
    parser.add_argument('--log', action='store_true',
                        help='show normalization logs (warning: for large files with many normalization rules this will'
                             ' cause a significant performance penalty and a lot of output data)')


def args_normalizers(parser: argparse.ArgumentParser):
    normalizers_desc = """
      A list of normalizers to execute on the input, can be one or more normalizers
      which are applied sequentially.
      The program will automatically find the normalizer in benchmarkstt.normalization.core,
      then benchmarkstt.normalization and finally in the global namespace.
      """

    normalizers = parser.add_argument_group('available normalizers', description=normalizers_desc)
    args_from_factory('normalizers', factory, normalizers)


def argparser(parser: argparse.ArgumentParser):
    """
    Adds the help and arguments specific to this module
    """
    args_logs(parser)

    files_desc = """
      You can provide multiple input and output files, each preceded by -i and -o
      respectively.
      If no input file is given, only one output file can be used.
      If using both multiple input and output files there should be an equal amount
      of each. Each processed input file will then be written to the corresponding
      output file."""

    files = parser.add_argument_group('input and output files', description=files_desc)
    args_inputfile(files)
    files.add_argument('-o', '--outputfile', action='append', nargs=1,
                       help='write output to this file, defaults to STDOUT',
                       metavar='file')

    args_normalizers(parser)
    return parser


def get_normalizer_from_args(args):
    if args.log:
        handler = logging.StreamHandler()
        formatter = DiffLoggingFormatter('cli', show_color_key=False)
        handler.setFormatter(formatter)
        handler.setLevel(logging.INFO)
        Logger.logger.addHandler(handler)

    composite = NormalizationComposite()

    if 'normalizers' in args:
        for item in args.normalizers:
            normalizer_name = item.pop(0).replace('-', '.')
            normalizer = factory.create(normalizer_name, *item)
            composite.add(normalizer)

    return composite


def main(parser, args):
    input_files = [f[0] for f in args.inputfile] if args.inputfile else None
    output_files = [f[0] for f in args.outputfile] if args.outputfile else None

    if 'normalizers' not in args or not len(args.normalizers):
        parser.error("need at least one normalizer")

    if input_files is None and output_files is not None:
        parser.error("can only write output to stdout when reading from stdin")

    composite = get_normalizer_from_args(args)

    encoding = settings.default_encoding
    if output_files is not None:
        # pre-open the output files before doing the grunt work
        output_files = [open(output_file, 'xt', encoding=encoding) for output_file in output_files]

    if input_files is not None:
        for idx, file in enumerate(input_files):
            with open(file, encoding=encoding) as input_file:
                text = input_file.read()
            text = composite.normalize(text)
            if output_files is None:
                sys.stdout.write(text)
            else:
                output_file = output_files[idx]
                output_file.write(text)
                output_file.close()
    else:
        text = sys.stdin.read()
        text = composite.normalize(text)
        sys.stdout.write(text)



from benchmarkstt import output
from benchmarkstt.schema import Schema
from collections import OrderedDict


class SimpleTextBase(output.Base):
    def print(self, result):
        if hasattr(result, '_asdict'):
            result = result._asdict()

        if type(result) is float:
            print("%.6f" % (result,))
        elif type(result) is dict or type(result) is OrderedDict:
            for k, v in result.items():
                print("%s: %r" % (k, v))
        else:
            print(result)


class ReStructuredText(SimpleTextBase):
    def result(self, title, result):
        print(title)
        print('=' * len(title))
        print()
        self.print(result)
        print()


class MarkDown(SimpleTextBase):
    def result(self, title, result):
        print('# %s' % (title,))
        print()
        self.print(result)
        print()


class Json(output.Base):
    def __init__(self):
        self._line = None

    def __enter__(self):
        if self._line is not None:
            raise ValueError("Already open")
        print('[')
        self._line = 0
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._line = None
        print('\n]')

    def result(self, title, result):
        if self._line != 0:
            print(',')
        self._line += 1
        print('\t', end='')

        if isinstance(result, tuple) and hasattr(result, '_asdict'):
            result = result._asdict()

        print(Schema.dumps(OrderedDict((('title', title), ('result', result)))), end='')



from difflib import SequenceMatcher
from benchmarkstt.diff import Base


class RatcliffObershelp(SequenceMatcher, Base):
    def __init__(self, a, b, *args, **kwargs):
        if 'autojunk' not in kwargs:
            kwargs['autojunk'] = False
        super().__init__(a=a, b=b, *args, **kwargs)



"""
Package benchmarkstt
"""

from .__meta__ import __author__, __version__
from functools import partial, wraps
from os import getenv


class DeferredCallback:
    """Simple helper class to defer the execution of formatting functions until it is needed"""

    def __init__(self, cb, *args, **kwargs):
        self._cb = wraps(cb)(partial(cb, *args, **kwargs))

    def __str__(self):
        return self._cb()

    def __repr__(self):
        return '<%s:%s>' % (self.__class__.__name__, repr(self._cb()))


class DeferredList:
    def __init__(self, cb):
        self._cb = cb
        self._list = None

    @property
    def list(self):
        if self._list is None:
            self._list = self._cb()
        return self._list

    def __getitem__(self, item):
        return self.list[item]


def make_printable(char):
    """
    Return printable representation of ascii/utf-8 control characters

    :param char:
    :return str:
    """
    if not len(char):
        return ''
    if len(char) > 1:
        return ''.join(list(map(make_printable, char)))

    codepoint = ord(char)
    if 0x00 <= codepoint <= 0x1f or 0x7f <= codepoint <= 0x9f:
        return chr(0x2400 | codepoint)

    return char if char != ' ' else '·'


class _Settings:
    @property
    def default_encoding(self):
        return getenv('DEFAULT_ENCODING', 'UTF-8')


settings = _Settings()



from benchmarkstt.factory import Factory


class Base:
    def __iter__(self):
        """
        Each segmentation class should be accessible as iterator, each iteration should
        return a Item, so the input format is essentially usable and can be easily
        converted to a :py:class:`benchmarkstt.schema.Schema`
        """
        raise NotImplementedError()


factory = Factory(Base)



import logging
from benchmarkstt import DeferredCallback


def log_call(logger: logging.Logger, log_level=None, result=None):
    """
    Decorator to log all calls to decorated function to given logger

    >>> import logging, sys, io
    >>>
    >>> logger = logging.getLogger('logger_name')
    >>> logger.setLevel(logging.DEBUG)
    >>> ch = logging.StreamHandler(sys.stdout)
    >>> ch.setFormatter(logging.Formatter('%(levelname)s:%(name)s: %(message)s'))
    >>> logger.addHandler(ch)
    >>>
    >>> @log_call(logger, logging.WARNING)
    ... def test(*args, **kwargs):
    ...     return 'result'
    >>> test('arg1', arg2='someval', arg3='someotherval')
    WARNING:logger_name: test('arg1', arg2='someval', arg3='someotherval')
    'result'
    >>> @log_call(logger, result=True)
    ... def test(*args, **kwargs):
    ...     return 'result'
    >>> test(arg2='someval', arg3='someotherval')
    DEBUG:logger_name: test(arg2='someval', arg3='someotherval')
    DEBUG:logger_name: test returned: result
    'result'
    """

    if log_level is None:
        log_level = logging.DEBUG

    def _log_call(func: callable):
        def _(*args, **kwargs):
            arguments_format = []
            arguments_list = []
            if len(args):
                arguments_format.append('%s')
                arguments_list.append(DeferredCallback(lambda: ', '.join([repr(a) for a in args])))
            if len(kwargs):
                arguments_format.append('%s')
                arguments_list.append(DeferredCallback(lambda: ', '.join([k + '=' + repr(kwargs[k]) for k in kwargs])))

            arguments_format = '%s(%s)' % (func.__name__, ', '.join(arguments_format))

            logger.log(log_level, arguments_format, *arguments_list)
            result_ = func(*args, **kwargs)
            if result:
                logger.log(log_level, '%s returned: %s', func.__name__, result_)
            return result_
        return _
    return _log_call



import logging
import os
from benchmarkstt.diff.formatter import DiffFormatter
from collections import namedtuple
from collections import OrderedDict

NormalizedLogItem = namedtuple('NormalizedLogItem', ['stack', 'original', 'normalized'])


class Logger:
    title = None
    logger = logging.getLogger('benchmarkstt.normalize')
    logger.setLevel(logging.INFO)
    logger.propagate = False
    stack = []


class ListHandler(logging.StreamHandler):
    def __init__(self):
        self._logs = []
        super().__init__(os.devnull)

    def emit(self, record):
        msg = self.format(record)
        self._logs.append(msg)

    @property
    def logs(self):
        return self._logs

    def flush(self):
        result = self._logs
        self._logs = []
        return result


class DiffLoggingFormatterDialect:
    def format(self, title, stack, diff):
        raise NotImplementedError()


class DiffLoggingTextFormatterDialect(DiffLoggingFormatterDialect):
    def format(self, title, stack, diff):
        args = []
        if title is not None:
            args.append(title)
        elif Logger.title is not None:
            args.append(Logger.title)
        args.append('/'.join(stack))
        args.append(diff)
        return ': '.join(args)


class DiffLoggingDictFormatterDialect(DiffLoggingFormatterDialect):
    def format(self, title, stack, diff):
        return OrderedDict(title=title, stack=stack, diff=diff)


class DiffLoggingFormatter(logging.Formatter):
    diff_logging_formatter_dialects = {
        "text": DiffLoggingTextFormatterDialect,
        "dict": DiffLoggingDictFormatterDialect,
    }

    def __init__(self, dialect=None, diff_formatter_dialect=None, title=None, *args, **kwargs):
        self._title = title
        self._differ = DiffFormatter(dialect, *args, **kwargs)
        strict = False
        if diff_formatter_dialect is None:
            diff_formatter_dialect = dialect
        else:
            strict = True

        self._formatter_dialect = self.get_dialect(diff_formatter_dialect, strict)
        super().__init__()

    def format(self, record):
        item = record.msg
        if type(item) is NormalizedLogItem:
            diff = self._differ.diff(item.original, item.normalized)
            return self._formatter_dialect.format(self._title, item.stack, diff)
        return super().format(record)

    @classmethod
    def has_dialect(cls, dialect):
        return dialect in cls.diff_logging_formatter_dialects

    @classmethod
    def get_dialect(cls, dialect, strict=None):
        if dialect is None:
            dialect = 'text'

        if not cls.has_dialect(dialect):
            if strict:
                raise ValueError("Unknown diff formatter dialect", dialect)

            dialect = 'text'

        return cls.diff_logging_formatter_dialects[dialect]()


def log(func):
    """
    Log decorator for normalization classes
    """

    def _(cls, text):
        Logger.stack.append(repr(cls))

        result = func(cls, text)
        logger_ = Logger.logger

        if text != result:
            logger_.info(NormalizedLogItem(list(Logger.stack), text, result))

        Logger.stack.pop()
        return result
    return _


class LogCapturer:
    def __init__(self, *args, **kwargs):
        self.formatter_args = (args, kwargs)
        self.handler = None

    def __enter__(self):
        self.handler = ListHandler()
        self.handler.setFormatter(DiffLoggingFormatter(*self.formatter_args[0], **self.formatter_args[1]))
        Logger.logger.addHandler(self.handler)
        return self

    @property
    def logs(self):
        return list(self.handler.logs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.handler.flush()
        Logger.logger.removeHandler(self.handler)
        self.handler = None



import textwrap
import inspect
import re
import ast
from collections import namedtuple
import logging
from docutils.core import publish_string
from docutils.writers import html5_polyglot
import docutils


logger = logging.getLogger(__name__)

Docblock = namedtuple('Docblock', ['docs', 'params', 'result', 'result_type'])
Param = namedtuple('Param', ['name', 'type', 'type_doc', 'is_required',
                             'description', 'examples'])
DocblockParam = namedtuple('DocblockParam', ['name', 'type', 'value'])


def format_docs(docs):
    return textwrap.dedent(docs).strip()


def doc_param_parser(docstring, key, no_name=None, allow_multiple=None,
                     replace_strat=None):
    results = [] if no_name or allow_multiple else {}

    if replace_strat is None:
        def replace_strat(match, param):
            return match.group(0)
    elif type(replace_strat) is str:
        _replace_strat = replace_strat

        def replace_strat(match, param):
            nonlocal _replace_strat
            return _replace_strat

    def _(match):
        nonlocal results, key, no_name, replace_strat
        if no_name:
            param = dict(name=key, type=match.group(1), value=match.group(2))
            return_val = replace_strat(match, param)
            results.append(DocblockParam(**param))
        else:
            value = textwrap.dedent(match.group(3)).strip()
            param = dict(name=match.group(2), type=match.group(1), value=value)
            return_val = replace_strat(match, param)
            param = DocblockParam(**param)
            if allow_multiple:
                # check if it already exists, if not create a new object
                idx = [idx for idx, val in enumerate(results)
                       if match.group(2) not in val]
                if not len(idx):
                    idx = len(results)
                    results.append({})
                else:
                    idx = idx[0]
                results[idx][match.group(2)] = param
            else:
                results[match.group(2)] = param

        return return_val

    if no_name:
        regex = r'^[ \t]*:%s[ \t]*([a-z_]+)?:[ \t]+(.*)$'
    else:
        regex = r'^[ \t]*:%s[ \t]+(?:([^:]+)[ \t]+)?([a-z_]+):[ \t]*(.+$|(?:$[ \t]*\n)*([ \t]+)([^\n]*)$(?:\4.*\n|\n)+)'

    docs = re.sub(
        regex % (re.escape(key),), _, docstring, flags=re.MULTILINE
    ).strip()

    return docs, results


def decode_literal(txt: str):
    if txt is None:
        return ''

    try:
        return ast.literal_eval(txt)
    except (ValueError, SyntaxError) as e:
        logger.warning('%s "%s" for: %s', type(e), e, txt)
        return txt


def parse(func):
    docs = format_docs(func.__doc__)

    argspec = inspect.getfullargspec(func)
    args = list(argspec.args)

    if len(args) and args[0] in ('self', 'cls'):
        args.pop(0)

    defaults_idx = len(args) - (len(argspec.defaults) if argspec.defaults else 0)

    docs, doc_params = doc_param_parser(docs, 'param', replace_strat='')
    docs, doc_result = doc_param_parser(docs, 'return', no_name=True)

    def decode_examples(match, param):
        if match.group(5) is None:
            param['value'] = decode_literal(param['value'])
        else:
            param['value'] = process_rst(param['value'], 'text')
        return ''

    docs, examples = doc_param_parser(docs, 'example', allow_multiple=True,
                                      replace_strat=decode_examples)

    params = []
    for idx, name in enumerate(args):
        type_ = None
        description = ''
        if name in doc_params:
            type_ = doc_params[name].type
            description = doc_params[name].value

        param = Param(name,
                      argspec.annotations[name] if name in argspec.annotations
                      else None,
                      type_,
                      idx < defaults_idx,
                      description,
                      examples)

        params.append(param)

    # quick hack to remove this
    docs = docs.replace(':py:class:', '')

    result = Docblock(docs=docs, params=params,
                      result=doc_result[0].value if doc_result else None,
                      result_type=doc_result[0].type if doc_result else None)
    return result


class HTML5Writer(html5_polyglot.Writer):
    def apply_template(self):
        subs = self.interpolation_dict()
        return subs['body']


class TextWriter(docutils.writers.Writer):
    class TextVisitor(docutils.nodes.SparseNodeVisitor):
        _text = ''

        def visit_Text(self, node):
            self._text += node.astext()

        def visit_paragraph(self, node):
            self._text += '\n\n'

        def text(self):
            return self._text

    def translate(self):
        visitor = self.TextVisitor(self.document)
        self.document.walkabout(visitor)
        self.output = visitor.text()


def process_rst(text, writer=None):
    if writer is None or writer == 'html':
        writer = HTML5Writer()
    elif writer == 'text':
        writer = TextWriter()
    elif type(writer) is str:
        raise ValueError("Unknown writer %s", str)
    settings = {'output_encoding': 'unicode', 'table_style': 'table'}
    return publish_string(text, writer=writer, writer_name='html5',
                          settings_overrides=settings)



import benchmarkstt.metrics as metrics
from io import StringIO
from benchmarkstt.input.core import PlainText
from benchmarkstt.normalization.core import Config
from benchmarkstt.normalization.logger import LogCapturer

factory = metrics.factory


def callback(cls, ref: str, hyp: str, config: str = None, return_logs: bool = None, *args, **kwargs):
    """
    :param ref: Reference text
    :param hyp: Hypothesis text
    :param config: The config to use
    :param bool return_logs: Return normalization logs

    :example ref: 'Hello darkness my OLD friend'
    :example hyp: 'Hello darkness my old foe'
    :example config:

            .. code-block:: text

                [normalization]
                # using a simple config file
                Lowercase

    :example result: ""
    """

    normalizer = None
    if config is not None and len(config.strip()):
        normalizer = Config(StringIO(config), section='normalization')

    ref = PlainText(ref, normalizer=normalizer)
    hyp = PlainText(hyp, normalizer=normalizer)

    metric = cls(*args, **kwargs)
    cls_name = cls.__name__.lower()

    if not return_logs:
        result = metric.compare(list(ref), list(hyp))
        if isinstance(result, tuple) and hasattr(result, '_asdict'):
            result = result._asdict()
        return {
            cls_name: result
        }

    with LogCapturer(dialect='html', diff_formatter_dialect='dict', title='Reference') as logcap:
        ref = list(ref)
        logs_ref = logcap.logs

    with LogCapturer(dialect='html', diff_formatter_dialect='dict', title='Hypothesis') as logcap:
        hyp = list(hyp)
        logs_hyp = logcap.logs

    result = metric.compare(ref, hyp)
    if isinstance(result, tuple) and hasattr(result, '_asdict'):
        result = result._asdict()

    return {
        cls_name: result,
        "logs": logs_ref + logs_hyp
    }



from benchmarkstt.schema import Schema
from benchmarkstt.factory import Factory


class Base:
    """
    Base class for metrics
    """
    def compare(self, ref: Schema, hyp: Schema):
        raise NotImplementedError()


factory = Factory(Base)



from benchmarkstt.normalization.logger import log
import logging
from benchmarkstt.factory import Factory
from benchmarkstt import settings
from benchmarkstt import csv
import os

_normalizer_namespaces = (
    "benchmarkstt.normalization.core",
    ""
)


logger = logging.getLogger(__name__)


class Base:
    @log
    def normalize(self, text: str) -> str:
        """
        Returns normalized text with rules supplied by the called class.
        """
        return self._normalize(text)

    def __repr__(self):
        return type(self).__name__

    def _normalize(self, text: str) -> str:
        raise NotImplementedError()


class BaseWithFileSupport(Base):
    """
    This kind of normalization class supports loading the values from a file, i.e.
    being wrapped in a core.File wrapper.
    """

    def _normalize(self, text: str) -> str:
        raise NotImplementedError()


class NormalizationComposite(Base):
    """
    Combining normalizers
    """

    def __init__(self, title=None):
        self._normalizers = []
        self._title = type(self).__name__ if title is None else title

    def add(self, normalizer):
        """Adds a normalizer to the composite "stack"
        """
        self._normalizers.append(normalizer)

    def _normalize(self, text: str) -> str:
        # allow for an empty file
        if not self._normalizers:
            return text

        for normalizer in self._normalizers:
            text = normalizer.normalize(text)
        return text

    def __repr__(self):
        return self._title


class File(Base):
    """
    Read one per line and pass it to the given normalizer

    :param str|class normalizer: Normalizer name (or class)
    :param file: The file to read rules from
    :param encoding: The file encoding

    :example text: "This is an Ex-Parakeet"
    :example normalizer: "regex"
    :example file: "./resources/test/normalizers/regex/en_US"
    :example encoding: "UTF-8"
    :example return: "This is an Ex Parrot"
    """

    def __init__(self, normalizer, file, encoding=None, path=None):
        if encoding is None:
            encoding = settings.default_encoding

        title = file
        if path is not None:
            file = os.path.join(path, file)

        with open(file, encoding=encoding) as f:
            self._normalizer = NormalizationComposite(title=title)
            for line in csv.reader(f):
                try:
                    self._normalizer.add(normalizer(*line))
                except TypeError as e:
                    raise ValueError("%s:%d %r(%r) %r" % (file, line.lineno, normalizer, line, e))

    def _normalize(self, text: str) -> str:
        return self._normalizer.normalize(text)


factory = Factory(Base, _normalizer_namespaces)


class FileFactory(Factory):
    def create(self, name, file=None, encoding=None, path=None):
        cls = super().__getitem__(name)
        return File(cls, file, encoding, path=path)

    def __getitem__(self, item):
        raise NotImplementedError("Not supported")


file_factory = FileFactory(BaseWithFileSupport, _normalizer_namespaces)



"""
Defines the main schema for comparison and implements json serialization
"""
import json
from collections.abc import Mapping
from typing import Union
from collections import defaultdict


class SchemaError(ValueError):
    """Top Error class for all schema related exceptions"""


class SchemaJSONError(SchemaError):
    """When loading incompatible JSON"""


class SchemaInvalidItemError(SchemaError):
    """Attempting to add an invalid item"""


class Item(Mapping):
    """
    Basic structure of each field to compare

    :raises: ValueError, SchemaInvalidItemError
    """

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            raise ValueError('Expected max 1 argument')
        if len(args) and len(kwargs):
            raise ValueError("Cannot combine both a positional and keyword arguments")
        if len(args):
            if not isinstance(args[0], dict):
                raise SchemaInvalidItemError("Expected a dict object", args[0])
            self._val = args[0]
        else:
            self._val = dict(**kwargs)
        self.meta = Meta()

    def __getitem__(self, k):
        return self._val[k]

    def __len__(self) -> int:
        return len(self._val)

    def __iter__(self):
        return iter(self._val)

    def __repr__(self):
        return 'Item(%s)' % (self.json(),)

    def json(self, **kwargs):
        return Schema.dumps(self, **kwargs)

    def _asdict(self):
        return self._val

    def __eq__(self, other):
        if type(other) is Item:
            other = other._asdict()
        return self._val == other

    def __ne__(self, other):
        return self._val != other


class Meta(defaultdict):
    """Containing metadata for an item, such as skipped"""


class Schema:
    """
    Basically a list of :py:class:`Item`
    """

    def __init__(self, data=None):
        # make Schema.dump/dumps methods available as instance methods
        self.dump = self.__dump
        self.dumps = self.__dumps
        if data is None:
            self._data = []
        else:
            self._data = [item if type(item) is Item else Item(item) for item in data]

    def __repr__(self):
        return 'Schema(%s)' % (self.json(),)

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, item):
        return self._data[item]

    """
    :raises: SchemaJSONError
    """
    @staticmethod
    def load(*args, **kwargs):
        return json.load(*args, **kwargs, cls=JSONDecoder)

    """
    :raises: SchemaJSONError
    """
    @staticmethod
    def loads(*args, **kwargs):
        return json.loads(*args, **kwargs, cls=JSONDecoder)

    @staticmethod
    def dump(cls, *args, **kwargs):
        return json.dump(cls, *args, **kwargs, cls=JSONEncoder)

    @staticmethod
    def dumps(cls, *args, **kwargs):
        return json.dumps(cls, *args, **kwargs, cls=JSONEncoder)

    def __dump(self, *args, **kwargs):
        return Schema.dump(self, *args, **kwargs)

    def __dumps(self, *args, **kwargs):
        return Schema.dumps(self, *args, **kwargs)

    def json(self, **kwargs):
        return self.dumps(**kwargs)

    def append(self, obj: Union[Item, dict]):
        if isinstance(obj, dict):
            obj = Item(obj)
        elif type(obj) is not Item:
            raise SchemaError("Wrong type", type(obj))
        self._data.append(obj)

    def extend(self, iterable):
        self._data.extend((item if type(item) is Item else Item(item) for item in iterable))

    def _aslist(self):
        return self._data

    def __eq__(self, other):
        if type(other) is Schema:
            other = other._aslist()
        return self._data == other

    def __ne__(self, other):
        return self._data != other


class JSONEncoder(json.JSONEncoder):
    """Custom JSON encoding for schema"""

    def default(self, o):
        if isinstance(o, Schema):
            return o._aslist()

        if isinstance(o, (tuple, Item)) and hasattr(o, '_asdict'):
            return o._asdict()

        return super().default(o)

    def encode(self, obj):
        try:
            obj = self.default(obj)
        except TypeError:
            pass

        return super().encode(obj)


class JSONDecoder(json.JSONDecoder):
    """Custom JSON decoding for schema"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, object_hook=self.object_hook)

    def decode(self, *args, **kwargs):
        result = super().decode(*args, **kwargs)
        if type(result) is not list:
            raise SchemaJSONError("Expected a list")
        return Schema(result)

    @staticmethod
    def object_hook(obj):
        return Item(obj)



"""
Make benchmarkstt available through a rudimentary JSON-RPC_ interface

.. attention::

    Only supported for Python versions 3.6 and above

.. _JSON-RPC: https://www.jsonrpc.org

"""

import jsonrpcserver
from flask import Flask, request, Response, render_template
from benchmarkstt.docblock import format_docs, parse, process_rst
from .jsonrpc import get_methods


def argparser(parser):
    """
    Adds the help and arguments specific to this module
    """

    parser.add_argument('--debug', action='store_true',
                        help='Run in debug mode')
    parser.add_argument('--host',
                        help='Hostname or ip to serve api')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port used by the server')
    parser.add_argument('--entrypoint', default='/api',
                        help='The jsonrpc api address')
    parser.add_argument('--list-methods', action='store_true',
                        help='List the available jsonrpc methods')
    parser.add_argument('--with-explorer', action='store_true',
                        help='Also create the explorer to test api calls with, '
                             'this is a rudimentary feature currently '
                             'only meant for testing and debugging.\n'
                             'Warning: the API explorer is provided as-is, without any tests '
                             'or code reviews. This is marked as a low-priority feature.')
    return parser


def create_app(entrypoint: str = None, with_explorer: bool = None):
    """
    Create the Flask app

    :param entrypoint: The HTTP path on which the api will be served
    :param bool with_explorer: Whether to also serve the JSON-RPC API explorer
    :return:
    """

    app = Flask(__name__)

    if entrypoint is None:
        entrypoint = '/api'

    methods = get_methods()

    @app.route(entrypoint, methods=["POST"])
    def jsonrpc():
        req = request.get_data().decode()
        response = jsonrpcserver.dispatch(req, methods=methods, debug=True, convert_camel_case=False)
        response_str = str(response)
        return Response(response_str, response.http_status, mimetype="application/json")

    if with_explorer:  # pragma: nocover
        app.template_filter('parse_rst')(process_rst)

        @app.route(entrypoint, methods=['GET'])
        def explorer():
            split_methods = {}
            individual_methods = []
            for name, func in methods.items.items():
                id_ = name.replace('.', '-')
                cat = 'METHODS'
                splitted = name.split('.', 1)
                if len(splitted) == 2:
                    cat = splitted[0]
                item = dict(id=id_, name=name, details=parse(func))
                individual_methods.append(item)

                if cat not in split_methods:
                    split_methods[cat] = []

                split_methods[cat].append(item)
            return render_template('api-explorer.html', grouped_methods=split_methods, methods=individual_methods)

    return app


def main(parser, args):  # pragma: nocover
    if args.list_methods:
        methods = get_methods()
        for name, func in methods.items.items():
            print('%s\n%s' % (name, '-' * len(name)))
            print('')
            print(format_docs(func.__doc__))
            print('')
    else:
        app = create_app(args.entrypoint, args.with_explorer)
        app.run(host=args.host, port=args.port, debug=args.debug)



"""
Some basic/simple normalization classes

"""

import re
import os
from unidecode import unidecode
from benchmarkstt import normalization
from benchmarkstt import config, settings
from contextlib import contextmanager
# from benchmarkstt.modules import LoadObjectProxy


file_types = (str,)
if hasattr(os, 'PathLike'):
    file_types = (str, os.PathLike)


class Replace(normalization.BaseWithFileSupport):
    """
    Simple search replace

    :param search: Text to search for
    :param replace: Text to replace with

    :example text: "Nudge nudge!"
    :example search: "nudge"
    :example replace: "wink"
    :example return: "Nudge wink!"
    """

    def __init__(self, search: str, replace: str):
        self._search = search
        self._replace = replace

    def _normalize(self, text: str) -> str:
        return text.replace(self._search, self._replace)


class ReplaceWords(normalization.BaseWithFileSupport):
    """
    Simple search replace that only replaces "words", the first letter will be
    checked case insensitive as well with preservation of case..

    :param search: Word to search for
    :param replace: Replace with

    :example text: "She has a heart of formica"
    :example search: "a"
    :example replace: "the"
    :example return: "She has the heart of formica"
    """

    def __init__(self, search: str, replace: str):
        search = search.strip()
        replace = replace.strip()

        args = tuple(map(re.escape, [
            search[0].upper(),
            search[0].lower(),
            search[1:] if len(search) > 1 else ''
        ]))
        regex = r'(?<!\w)[%s%s]%s(?!\w)' % args
        self._pattern = re.compile(regex)
        self._replace = replace

    def _replacement_callback(self, matches):
        if len(self._replace) == 0:
            return ''

        if matches.group(0)[0].isupper():
            return ''.join([self._replace[0].upper(), self._replace[1:]])

        return ''.join([self._replace[0].lower(), self._replace[1:]])

    def _normalize(self, text: str) -> str:
        return self._pattern.sub(self._replacement_callback, text)


class Regex(normalization.BaseWithFileSupport):
    r"""
    Simple regex replace. By default the pattern is interpreted
    case-sensitive.

    Case-insensitivity is supported by adding inline modifiers.

    You might want to use capturing groups to preserve the case. When replacing
    a character not captured, the information about its case is lost...

    Eg. would replace "HAHA! Hahaha!" to "HeHe! Hehehe!":

     +------------------+-------------+
     | search           | replace     |
     +==================+=============+
     | ``(?i)(h)a``     | ``\1e``     |
     +------------------+-------------+


    No regex flags are set by default, you can set them yourself though in the
    regex, and combine them at will, eg. multiline, dotall and ignorecase.

    Eg. would replace "New<CRLF>line" to "newline":

     +------------------------+------------------+
     | search                 | replace          |
     +========================+==================+
     | ``(?msi)new.line``     | ``newline``      |
     +------------------------+------------------+

    :example text: "HAHA! Hahaha!"
    :example search: '(?i)(h)a'
    :example replace: '\\1e'
    :example return: "HeHe! Hehehe!"
    """

    def __init__(self, search: str, replace: str):
        self._pattern = re.compile(search)
        self._substitution = replace

    def _normalize(self, text: str) -> str:
        return self._pattern.sub(self._substitution, text)


class Lowercase(normalization.Base):
    """
    Lowercase the text


    :example text: "Easy, Mungo, easy... Mungo..."
    :example return: "easy, mungo, easy... mungo..."
    """

    def _normalize(self, text: str) -> str:
        return text.lower()


class Unidecode(normalization.Base):
    """
    Unidecode characters to ASCII form, see `Python's Unidecode package
    <https://pypi.org/project/Unidecode>`_ for more info.

    :example text: "𝖂𝖊𝖓𝖓 𝖎𝖘𝖙 𝖉𝖆𝖘 𝕹𝖚𝖓𝖘𝖙ü𝖈𝖐 𝖌𝖎𝖙 𝖚𝖓𝖉 𝕾𝖑𝖔𝖙𝖊𝖗𝖒𝖊𝖞𝖊𝖗?"
    :example return: "Wenn ist das Nunstuck git und Slotermeyer?"
    """

    def _normalize(self, text: str) -> str:
        return unidecode(text)


class ConfigSectionNotFoundError(ValueError):
    """
    Raised when a requested config section was not found
    """


class Config(normalization.Base):
    doc_string = r"""
    Use config file notation to define normalization rules. This notation is a
    list of normalizers, one per line.

    Each normalizer that is based needs a file is followed by a file name of a
    csv, and can be optionally followed by the file encoding (if different than
    default).
    All options are loaded in from this csv and applied to the normalizer.

    The normalizers can be any of the core normalizers, or you can refer to your
    own normalizer class (like you would use in a python import, eg.
    `my.own.package.MyNormalizerClass`).

    Additional rules:
      - Normalizer names are case-insensitive.
      - Arguments MAY be wrapped in double quotes.
      - If an argument contains a space, newline or double quote, it MUST be
        wrapped in double quotes.
      - A double quote itself is represented in this quoted argument as two
        double quotes: ``""``.

    The normalization rules are applied top-to-bottom and follow this format::

        {[section]}
        # This is a comment

        # (Normalizer2 has no arguments)
        lowercase

        # loads regex expressions from regexrules.csv in "utf 8" encoding
        regex regexrules.csv "utf 8"

        # load another config file, [section1] and [section2]
        config configfile.ini section1
        config configfile.ini section2

        # loads replace expressions from replaces.csv in default encoding
        replace     replaces.csv

    :param file: The config file
    :param encoding: The file encoding
    :param section: The subsection of the config file to use, {section}

    :example text: "He bravely turned his tail and fled"
    :example file: "./resources/test/normalizers/configfile.conf"
    :example encoding: "UTF-8"
    :example return: "ha bravalY Turnad his tail and flad"
    """

    MAIN_SECTION = object()
    _default_section = 'normalization'

    def __init__(self, file, section=None, encoding=None):
        if encoding is None or encoding == '':
            encoding = settings.default_encoding

        if section is None:
            section = self._default_section
        elif section is self.MAIN_SECTION:
            section = None

        if type(file) in file_types:
            # next filenames are relative from path of the config file...
            path = os.path.dirname(os.path.realpath(file))
            title = file

            with open(file, encoding=encoding) as f:
                reader = config.reader(f)
        else:
            path = None
            title = ''
            reader = config.reader(file)

        if section is not None:
            if section not in reader:
                raise ConfigSectionNotFoundError(section)

            reader = reader[section]
            title += '[%s]' % (section,)

        self._normalizer = normalization.NormalizationComposite(title)

        for line in reader:
            try:
                if line[0] in normalization.file_factory:
                    normalizer = normalization.file_factory.create(*line, path=path)
                else:
                    normalizer = normalization.factory.create(*line)
                self._normalizer.add(normalizer)
            except ImportError:
                raise ValueError("Unknown normalizer %s on line %d: %s" %
                                 (repr(line[0]), line.lineno, repr(' '.join(line))))

    def _normalize(self, text: str) -> str:
        return self._normalizer.normalize(text)

    @classmethod
    @contextmanager
    def default_section(cls, section):
        prev_section = cls._default_section
        cls._default_section = section
        try:
            cls.refresh_docstring()
            yield
        finally:
            cls._default_section = prev_section

    @classmethod
    def refresh_docstring(cls):
        section = 'defaults to %s' % (repr(cls._default_section),) if cls._default_section else 'no section by default'
        section_tag = '[%s]' % (cls._default_section,) if cls._default_section else ''
        cls.__doc__ = cls.doc_string.replace('{section}', section).replace('{[section]}', section_tag)


# For future versions
# class ExternalNormalizer(LoadObjectProxy, normalization.BaseWithFileSupport):
#     """
#     Automatically loads an external normalizer class.
#
#     :param name: The name of the normalizer to load (eg. mymodule.normalization.Normalizer)
#     """


Config.refresh_docstring()



from . import csv
import re


class SectionConfigReader:
    def __init__(self, config):
        self.config = list(config)
        sections = {}
        regex = re.compile(r'\[[a-z0-9]+\]', re.IGNORECASE)
        prev_idx = 0
        prev_section = None
        for idx, line in enumerate(self.config):
            if len(line) == 1 and regex.match(line[0]):
                sections[prev_section] = slice(prev_idx + 1 if prev_idx != 0 else 0, idx)
                prev_idx = idx
                prev_section = line[0][1:-1]

        sections[prev_section] = slice(prev_idx + 1, idx + 1)
        self.sections = sections

    def __iter__(self):
        return iter(self.config)

    def __getitem__(self, k):
        return self.config[self.sections[k]]

    def __contains__(self, item):
        return item in self.sections


def reader(file):
    csvreader = csv.reader(file, 'whitespace')
    sectionreader = SectionConfigReader(csvreader)
    return sectionreader



from benchmarkstt.schema import Schema
import logging
from benchmarkstt.diff.core import RatcliffObershelp
from benchmarkstt.diff.formatter import format_diff
from benchmarkstt.metrics import Base
from collections import namedtuple
# from benchmarkstt.modules import LoadObjectProxy

logger = logging.getLogger(__name__)

OpcodeCounts = namedtuple('OpcodeCounts',
                          ('equal', 'replace', 'insert', 'delete'))


def traversible(schema, key=None):
    if key is None:
        key = 'item'
    return [word[key] for word in schema]


def get_opcode_counts(opcodes):
    counts = OpcodeCounts(0, 0, 0, 0)._asdict()
    for tag, alo, ahi, blo, bhi in opcodes:
        if tag == 'equal':
            counts[tag] += ahi - alo
        elif tag == 'insert':
            counts[tag] += bhi - blo
        elif tag == 'delete':
            counts[tag] += ahi - alo
        elif tag == 'replace':
            ca = ahi - alo
            cb = bhi - blo
            if ca < cb:
                counts['insert'] += cb - ca
                counts['replace'] += ca
            elif ca > cb:
                counts['delete'] += ca - cb
                counts['replace'] += cb
            else:
                counts[tag] += ahi - alo
    return OpcodeCounts(counts['equal'], counts['replace'], counts['insert'], counts['delete'])


def get_differ(a, b, differ_class):
    if differ_class is None:
        # differ_class = HuntMcIlroy
        differ_class = RatcliffObershelp
    return differ_class(traversible(a), traversible(b))


class WordDiffs(Base):
    """
    Present differences on a per-word basis

    :param dialect: Presentation format. Default is 'cli'.
    :example dialect: 'html'
    :param differ_class: For future use.
    """

    def __init__(self, dialect=None, differ_class=None):
        self._differ_class = differ_class
        self._dialect = dialect

    def compare(self, ref: Schema, hyp: Schema):
        differ = get_differ(ref, hyp, differ_class=self._differ_class)
        a = traversible(ref)
        b = traversible(hyp)
        return format_diff(a, b, differ.get_opcodes(),
                           dialect=self._dialect,
                           preprocessor=lambda x: ' %s' % (' '.join(x),))


class WER(Base):
    """
    Word Error Rate, basically defined as::

        insertions + deletions + substitions
        ------------------------------------
             number of reference words

    See: https://en.wikipedia.org/wiki/Word_error_rate

    Insertions, deletions and substitutions are
    identified using the Hunt–McIlroy diff algorithm.
    This algorithm is the one used internally by Python.
    See https://docs.python.org/3/library/difflib.html

    :param mode: WER variant. 'strict' is the default. 'hunt' applies 0.5 weight to insertions and deletions.
    :param differ_class: For future use.
    """

    # WER modes
    MODE_STRICT = 'strict'
    MODE_HUNT = 'hunt'

    DEL_PENALTY = 1
    INS_PENALTY = 1
    SUB_PENALTY = 1

    def __init__(self, mode=None, differ_class=None):
        if differ_class is None:
            differ_class = RatcliffObershelp
        self._differ_class = differ_class
        if mode == self.MODE_HUNT:
            self.DEL_PENALTY = self.INS_PENALTY = .5

    def compare(self, ref: Schema, hyp: Schema):
        diffs = get_differ(ref, hyp, differ_class=self._differ_class)

        counts = get_opcode_counts(diffs.get_opcodes())

        changes = counts.replace * self.SUB_PENALTY + \
            counts.delete * self.DEL_PENALTY + \
            counts.insert * self.INS_PENALTY

        total_ref = counts.equal + counts.replace + counts.delete
        if total_ref == 0:
            return 1
        return changes / total_ref


class DiffCounts(Base):
    """
    Get the amount of differences between reference and hypothesis
    """

    def __init__(self, differ_class=None):
        if differ_class is None:
            differ_class = RatcliffObershelp
        self._differ_class = differ_class

    def compare(self, ref: Schema, hyp: Schema):
        diffs = get_differ(ref, hyp, differ_class=self._differ_class)
        return get_opcode_counts(diffs.get_opcodes())


# For a future version
# class ExternalMetric(LoadObjectProxy, Base):
#     """
#     Automatically loads an external metric class.
#
#     :param name: The name of the metric to load (eg. mymodule.metrics.MyOwnMetricClass)
#     """



"""
Calculate metrics based on the comparison of a hypothesis with a reference.
"""

from benchmarkstt.input import core
from benchmarkstt.output import factory as output_factory
from benchmarkstt.metrics import factory
from benchmarkstt.cli import args_from_factory
from benchmarkstt.normalization.logger import Logger
import argparse
from inspect import signature, Parameter
import logging
from collections import OrderedDict


def argparser(parser: argparse.ArgumentParser):
    # steps: input normalize[pre?] segmentation normalize[post?] compare

    parser.add_argument('-r', '--reference', help='File to use as reference', required=True)
    parser.add_argument('-h', '--hypothesis', help='File to use as hypothesis', required=True)

    types = OrderedDict(infer=' '.join([core.File.__doc__.strip(),
                                        'Automatically infer file type from the filename extension.']),
                        argument='Read the argument and treat as plain text (without reading from file)',
                        **core.File.available_types())
    types_help = ['Available types:']
    types_help.extend(['%r: %s' % (k, v) for k, v in types.items()])
    types_help = '\n'.join(types_help)

    subparser = parser.add_argument_group('reference and hypothesis types',
                                          description='You can specify which file type the --reference/-r and ' +
                                                      '--hypothesis/-h arguments should be treated as.\n\n' +
                                                      types_help)

    subparser.add_argument('-rt', '--reference-type', default='infer',
                           help='Type of reference file', choices=types.keys())
    subparser.add_argument('-ht', '--hypothesis-type', default='infer',
                           help='Type of hypothesis file', choices=types.keys())

    parser.add_argument('-o', '--output-format', default='restructuredtext', choices=output_factory.keys(),
                        help='Format of the outputted results')

    metrics_desc = "A list of metrics to calculate. At least one metric needs to be provided."

    subparser = parser.add_argument_group('available metrics', description=metrics_desc)
    args_from_factory('metrics', factory, subparser)
    return parser


def file_to_iterable(file, type_, normalizer=None):
    if type_ == 'argument':
        return core.PlainText(file, normalizer=normalizer)
    return core.File(file, type_, normalizer=normalizer)


def main(parser, args, normalizer=None):
    logging.getLogger()
    prev_title = Logger.title
    Logger.title = 'Reference'
    ref = list(file_to_iterable(args.reference, args.reference_type, normalizer=normalizer))
    Logger.title = 'Hypothesis'
    hyp = list(file_to_iterable(args.hypothesis, args.hypothesis_type, normalizer=normalizer))
    Logger.title = prev_title

    if 'metrics' not in args or not len(args.metrics):
        parser.error("need at least one metric")

    with output_factory.create(args.output_format) as out:
        for item in args.metrics:
            metric_name = item.pop(0).replace('-', '.')
            cls = factory[metric_name]
            kwargs = dict()

            # somewhat hacky default diff formats for metrics
            sig = signature(cls.__init__).parameters
            sigkeys = list(sig)

            if 'dialect' in sigkeys:
                idx = sigkeys.index('dialect') - 1
                sig = sig['dialect']
                if sig.kind in (Parameter.POSITIONAL_OR_KEYWORD, Parameter.POSITIONAL_ONLY):
                    if len(item) <= idx:
                        if args.output_format == 'json':
                            kwargs['dialect'] = 'list'
                            if 'diff_formatter_dialect' in sigkeys:
                                kwargs['diff_formatter_dialect'] = 'dict'
                        else:
                            kwargs['dialect'] = 'cli'

            metric = cls(*item, **kwargs)
            result = metric.compare(ref, hyp)
            out.result(metric_name, result)



class Registry:
    """
    Simple registry class holding aliases and their corresponding values
    """

    def __init__(self):
        self._registry = {}

    def __contains__(self, item):
        return item in self._registry

    def __getitem__(self, item):
        return self._registry[item]

    def __setitem__(self, key, value):
        if key in self._registry:
            raise ValueError("Conflict: alias '%s' is already registered" % (key,))
        self._registry[key] = value

    def __delitem__(self, key):
        del self._registry[key]

    def register(self, key, value):
        return self.__setitem__(key, value)

    def unregister(self, key):
        return self.__delitem__(key)

    def keys(self):
        return self._registry.keys()



"""
Module providing our own CSV file parser with support for whitespace trimming, empty lines filtering and comment lines
"""

import typing
import sys
from functools import partial
from benchmarkstt import DeferredList, make_printable


class InvalidDialectError(ValueError):
    """An invalid dialect was supplied"""


class UnknownDialectError(ValueError):
    """An unknown dialect was requested"""


class CSVParserError(ValueError):
    """Some error occured while attempting to parse the file"""

    def __init__(self, message, line, char, index):
        self.message = message
        self.line = line
        self.char = char
        self.index = index


class UnclosedQuoteError(CSVParserError):
    """A quote wasn't properly closed"""


class UnallowedQuoteError(CSVParserError):
    """A quote is not allowed there"""


class Dialect:
    delimiter = None
    quotechar = None
    commentchar = None
    trimleft = None
    trimright = None


class DefaultDialect(Dialect):
    delimiter = ','
    quotechar = '"'
    commentchar = '#'
    trimleft = ' \t\n\r'
    trimright = trimleft
    ignoreemptylines = True


class WhitespaceDialect(DefaultDialect):
    delimiter = ' \t'


known_dialects = {
    "default": DefaultDialect,
    "whitespace": WhitespaceDialect
}

MODE_FIRST = 0
MODE_OUTSIDE = 1
MODE_INSIDE = 2
MODE_INSIDE_QUOTED = 3
MODE_INSIDE_QUOTED_QUOTE = 4
MODE_COMMENT = 5

Line = list
Field = str


class Line(list):
    @property
    def lineno(self):
        return self.__dict__['lineno']

# don't really know if it's quoted atm
# class Field(str, object):
#     @property
#     def quoted(self):
#         return self.__dict__['quoted']
#
#     @quoted.setter
#     def quoted(self, value):
#         self.__dict__['quoted'] = bool(value)


class Reader:
    """
    CSV-like file reader with support for comment chars, ignoring empty lines
    and whitespace trimming on both sides of each field.

    """

    def __init__(self, file: typing.io.TextIO, dialect: Dialect, debug=None):
        if not issubclass(dialect, Dialect):
            raise InvalidDialectError("Invalid dialect", dialect)
        self.line_num = 0
        self._dialect = dialect
        self._file = file
        self._debug = bool(debug)

    def _trimright(self, data: str):
        chars = self._dialect.trimright
        if chars is None:
            return data
        return data.rstrip(chars)

    def _is_ignore_left(self, char: str):
        if self._dialect.trimleft is None:
            return False
        return char in self._dialect.trimleft

    def _is_ignore_right(self, char: str):
        if self._dialect.trimright is None:
            # currently no dialect with no trimright
            return False  # pragma: nocover
        return char in self._dialect.trimright

    def _is_comment(self, char: str):
        if self._dialect.commentchar is None:
            return False
        return char in self._dialect.commentchar

    def _is_quote(self, char: str):
        if self._dialect.quotechar is None:
            return False
        return char == self._dialect.quotechar

    def _is_delimiter(self, char: str):
        return char in self._dialect.delimiter

    def __iter__(self):
        readchar = iter(partial(self._file.read, 1), '')
        cur_line = 1

        if self._debug:
            current_module = sys.modules[__name__]
            # print the color key the different modes
            print('MODES: ', end='')
            print(' '.join(['\033[1;%d;40m%s\033[0;0m' % (32 + getattr(current_module, name), name[5:])
                            for name in dir(current_module)
                            if name.startswith('MODE_')
                            ]))

            def debug(txt='', args=tuple(), **kwargs):
                if type(args) is not tuple:
                    args = tuple(DeferredList(args))
                print(txt % args, **kwargs)
                pass
        else:
            def debug(*args, **kwargs):
                pass

        newlinechars = '\n\r'

        mode = MODE_FIRST
        field = []
        line = Line()

        if self._dialect.trimright is not None:
            delimiter_is_whitespace = self._dialect.delimiter in self._dialect.trimright
        else:
            delimiter_is_whitespace = False

        def yield_line():
            nonlocal line, field, mode, delimiter_is_whitespace, is_newline, cur_line
            if not(mode == MODE_OUTSIDE and delimiter_is_whitespace):
                next_field()
            field = []
            _line = line
            _line.__dict__['lineno'] = cur_line
            line = Line()
            mode = MODE_FIRST
            return _line

        def next_field():
            nonlocal field, line, mode
            field = ''.join(field)
            if mode != MODE_INSIDE_QUOTED_QUOTE:
                field = self._trimright(field)

            field = Field(field)
            line.append(field)

            field = []
            mode = MODE_OUTSIDE

        cur_char = 0
        last_quote_line = None
        last_quote_char = None
        last_quote_idx = None
        idx = 0
        for char in readchar:
            cur_char += 1
            idx += 1

            # print char to stdout with color defining mode
            debug('\033[1;%d;40m%s\033[0;0m', lambda: (32+mode, make_printable(char)), end='')

            is_newline = char in newlinechars
            if is_newline:
                cur_line += 1
                cur_char = 0

            if mode == MODE_COMMENT:
                if is_newline:
                    mode = MODE_FIRST
                continue

            if mode in (MODE_OUTSIDE, MODE_FIRST):
                if is_newline:
                    if mode != MODE_FIRST:
                        yield yield_line()
                    continue

                if self._is_ignore_left(char):
                    continue

                if self._is_comment(char):
                    if mode is MODE_OUTSIDE:
                        yield yield_line()
                    mode = MODE_COMMENT
                    continue

                if self._is_quote(char):
                    mode = MODE_INSIDE_QUOTED
                    last_quote_line = cur_line
                    last_quote_char = cur_char
                    last_quote_idx = idx
                    continue

                if self._is_delimiter(char):
                    next_field()
                    continue

                mode = MODE_INSIDE
                field.append(char)
                continue

            if mode == MODE_INSIDE:
                if self._is_quote(char):
                    raise UnallowedQuoteError("Quote not allowed here", cur_line, cur_char, idx)

                if is_newline:
                    yield yield_line()
                    continue

                if self._is_delimiter(char):
                    next_field()
                    continue

                field.append(char)
                continue

            if mode == MODE_INSIDE_QUOTED_QUOTE:
                if self._is_quote(char):
                    field.append(char)
                    mode = MODE_INSIDE_QUOTED
                    continue

                if self._is_delimiter(char):
                    next_field()
                    continue

                if is_newline:
                    yield yield_line()
                    continue

                if not delimiter_is_whitespace:
                    if self._is_ignore_right(char):
                        continue
                    if self._is_comment(char):
                        yield yield_line()
                        mode = MODE_COMMENT
                        continue

                raise UnallowedQuoteError("Single quote inside quoted field", cur_line, cur_char, idx)

            if mode == MODE_INSIDE_QUOTED:
                if self._is_quote(char):
                    mode = MODE_INSIDE_QUOTED_QUOTE
                    continue

                field.append(char)
                continue

        debug()

        if mode == MODE_INSIDE_QUOTED:
            raise UnclosedQuoteError("Unexpected end", last_quote_line, last_quote_char, last_quote_idx)

        if mode in (MODE_INSIDE_QUOTED_QUOTE, MODE_OUTSIDE, MODE_INSIDE):
            yield yield_line()


def reader(file: typing.io.TextIO, dialect: typing.Union[None, str, Dialect] = None, **kwargs) -> Reader:
    if dialect is None:
        dialect = DefaultDialect
    elif type(dialect) is str:
        if dialect not in known_dialects:
            raise UnknownDialectError("Dialect not known", dialect)
        dialect = known_dialects[dialect]

    return Reader(file, dialect, **kwargs)



"""
Make benchmarkstt available through a rudimentary JSON-RPC_ interface

.. warning::
    Only supported for Python versions 3.6 and above!

.. _JSON-RPC: https://www.jsonrpc.org

"""

import jsonrpcserver
import json
from benchmarkstt import __meta__
from functools import wraps
from benchmarkstt.docblock import format_docs
from benchmarkstt.modules import Modules
from inspect import _empty, Parameter, signature
import os


class SecurityError(ValueError):
    """Trying to do or access something that isn't allowed"""


class MagicMethods:
    possible_path_args = ['file', 'path']

    def __init__(self):
        self.methods = jsonrpcserver.methods.Methods()

    @staticmethod
    def is_safe_path(path):
        """
        Determines whether the file or path is within the current working directory
        :param str|PathLike path:
        :return: bool
        """
        return os.path.abspath(path).startswith(os.path.abspath(os.getcwd()))

    def serve(self, config, callback):
        """
        Responsible for creating a callback with proper documentation and arguments
        signature that can be registered as an api call.

        :param config:
        :param callback:
        :return: callable
        """
        cls = config.cls

        @wraps(cls)
        def _(*args, **kwargs):
            # only allow files from cwd to be used...
            try:
                # todo (?) add available files and folders as select options
                for name in self.possible_path_args:
                    if name in kwargs:
                        if not self.is_safe_path(kwargs[name]):
                            raise SecurityError("Access to unallowed file attempted", name)
            except SecurityError as e:
                data = {
                    "message": e.args[0],
                    "field": e.args[1]
                }
                raise AssertionError(json.dumps(data))

            result = callback(cls, *args, **kwargs)
            if isinstance(result, tuple) and hasattr(result, '_asdict'):
                result = result._asdict()
            return result

        # copy signature from original
        sig = signature(cls)

        def param_filter(param):
            return param.kind not in (Parameter.VAR_KEYWORD, Parameter.VAR_POSITIONAL)

        cb_params = signature(callback).parameters.values()
        cb_params = list(filter(param_filter, cb_params))

        extra_params = [parameter for parameter in cb_params
                        if parameter.name != 'cls']

        if len(extra_params):
            params = list(filter(lambda x: x.default is _empty, extra_params))
            params.extend(filter(param_filter, sig.parameters.values()))
            params.extend(list(filter(lambda x: x.default is not _empty, extra_params)))

        sig = sig.replace(parameters=params)

        _.__doc__ += callback.__doc__
        _.__signature__ = sig
        return _

    def load(self, name, module):
        """
        Load all possible callbacks for a given module

        :param name:
        :param Module module:
        """
        factory = module.factory
        callables = list(factory)

        def lister():
            """
            Get a list of available core %s

            :return object: With key being the %s name, and value its description
            """
            return {config.name: config.docs for config in callables}

        lister.__doc__ = lister.__doc__ % (name, name)

        self.register("list.%s" % (name,), lister)

        # add each callable as its own api call
        for conf in callables:
            apicallname = '%s.%s' % (name, conf.name,)
            self.register(apicallname, self.serve(conf, module.callback))

    def register(self, name, callback):
        """
        Register a callback as an api call
        :param name:
        :param callback:
        """
        self.methods.add(**{name: callback})


class DefaultMethods:
    @staticmethod
    def version():
        """
        Get the version of benchmarkstt

        :return str: BenchmarkSTT version
        """
        return __meta__.__version__

    @staticmethod
    def help(methods):
        def _():
            """
            Returns available api methods

            :return object: With key being the method name, and value its description
            """
            return {name: format_docs(func.__doc__) for name, func in methods.items.items()}
        return _


def get_methods() -> jsonrpcserver.methods.Methods:
    """
    Returns the available JSON-RPC api methods

    :return: jsonrpcserver.methods.Methods
    """

    methods = MagicMethods()
    methods.register('version', DefaultMethods.version)
    for name, module in Modules('api'):
        methods.load(name, module)

    methods.register('help', DefaultMethods.help(methods.methods))
    return methods.methods



from benchmarkstt.factory import Factory


class Base:
    def __init__(self, a='', b=''):
        raise NotImplementedError()

    def get_opcodes(self):
        """
        Return list of 5-tuples describing how to turn `a` into `b`.

        Each tuple is of the form `(tag, i1, i2, j1, j2)`. The first tuple has
        `i1 == j1 == 0`, and remaining tuples have `i1` equals the `i2` from the
        tuple preceding it, and likewise for `j1` equals the previous `j2`.

        The tags are strings, with these meanings:

         - 'replace': `a[i1:i2]` should be replaced by `b[j1:j2]`
         - 'delete': `a[i1:i2]` should be deleted. Note that `j1==j2` in this case.
         - 'insert': `b[j1:j2]` should be inserted at `a[i1:i1]`. Note that `i1==i2` in this case.
         - 'equal': `a[i1:i2] == b[j1:j2]`
        """
        raise NotImplementedError()


factory = Factory(Base)



"""
Default input formats

"""

import benchmarkstt.segmentation.core as segmenters
from benchmarkstt import input, settings
# from benchmarkstt.modules import LoadObjectProxy


class PlainText(input.Base):
    """
    plain text
    """
    def __init__(self, text, segmenter=None, normalizer=None):
        if segmenter is None:
            segmenter = segmenters.Simple
        self._text = text
        self._segmenter = segmenter
        self._normalizer = normalizer

    def __iter__(self):
        return iter(self._segmenter(self._text, normalizer=self._normalizer))


class File(input.Base):
    """
    Load from a given filename.
    """

    _extension_to_class = {
        "txt": PlainText,
    }

    @classmethod
    def available_types(cls):
        return {cls_config.name: ' '.join([cls.__doc__.strip(),
                                           'Treat file as',
                                           cls_config.cls.__doc__.strip()])
                for cls_config in input.factory
                if cls_config.name != 'file'}

    def __init__(self, file, input_type=None, normalizer=None):
        self._normalizer = normalizer
        if input_type is None or input_type == 'infer':
            if '.' not in file:
                raise ValueError('Cannot infer input file type of files without an extension')

            extension = file.rsplit('.', 1)[1].lower()
            if extension not in self._extension_to_class:
                raise ValueError('Cannot infer input file type for files of extension %s' % (extension,))

            input_type = self._extension_to_class[extension]

        encoding = settings.default_encoding
        with open(file, encoding=encoding):
            """Just checks that file is readable..."""

        self._file = file

        if type(input_type) is str:
            input_type = input.factory[input_type]

        self._input_class = input_type

    def __iter__(self):
        encoding = settings.default_encoding
        with open(self._file, encoding=encoding) as f:
            text = f.read()

        return iter(self._input_class(text, normalizer=self._normalizer))

# For future versions
# class ExternalInput(LoadObjectProxy, input.Base):
#     """
#     Automatically loads an external input class.
#
#     :param name: The name of the input to load (eg. mymodule.input.MyFileFormat)
#     """



import sys
from importlib import import_module

_modules = ['normalization', 'metrics', 'benchmark']

if sys.version_info >= (3, 6):
    # only supported in python >= 3.6
    _modules.append('api')


class Modules:
    def __init__(self, sub_module):
        self._postfix = '' if sub_module is None else '.' + sub_module

    def __iter__(self):
        for module in _modules:
            try:
                yield (module, self[module])
            except IndexError:
                pass

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, key):
        name = 'benchmarkstt.%s%s' % (key, self._postfix)
        try:
            module = import_module(name)
            if hasattr(module, 'hidden'):
                if module.hidden:
                    raise ImportError()
            return module
        except ImportError:
            raise IndexError('Module not found', key)

    def keys(self):
        return [key for key, value in iter(self)]


def load_object(name, transform=None):
    """
    Load an object based on a string.
    :param name: The string representation of an object
    :param transform: Transform (callable) done on the object name for comparison, if None, will lowercase compare.
    False for no transform.
    """
    module = list(name.split('.'))

    if transform is None:
        transform = str.lower
    elif transform is False:
        def identity(x):
            return x
        transform = identity

    class_name = transform(module.pop())
    module = '.'.join(module)
    if len(module) == 0:
        module = globals()
    else:
        module = import_module(module)

    for found_class_name in dir(module):
        if transform(found_class_name) != class_name:
            continue
        return getattr(module, found_class_name)

    raise ImportError("Could not find an object for %r" % (name,))


class Proxy:
    """
    Pass all function calls to instance, or to parent class if instance does not
    implement it.
    """

    def __init__(self, instance):
        self._instance = instance

    def __getattribute__(self, item):
        cls = object.__getattribute__(self, '_instance')

        if item == '_instance':
            return cls

        if hasattr(cls, item):
            return getattr(cls, item)

        return object.__getattribute__(self, item)


class LoadObjectProxy(Proxy):
    """
    Automatically load a class from any namespace, and pass all function calls to it,
    or to parent class if it is not implemented.
    """

    def __init__(self, name, *args, **kwargs):
        super().__init__(load_object(name)(*args, **kwargs))



import argparse
import logging
import textwrap
import itertools
from benchmarkstt.modules import Modules
from benchmarkstt import __meta__
from benchmarkstt.normalization.core import Config
from argparse import ArgumentError
from contextlib import contextmanager
import sys
from functools import partial
import re

# allow loading of modules based on current working directory
sys.path.insert(0, '')


def args_help(parser):
    parser.add_argument('--help', action='help', default=argparse.SUPPRESS,
                        help=argparse._('Show this help message and exit'))


def args_common(parser):
    parser.add_argument('--log-level', type=str.lower, default='warning', dest='log_level',
                        choices=list(map(str.lower, logging._nameToLevel.keys())),
                        help=argparse._('Set the logging output level'))


def args_from_factory(action, factory, parser):
    for conf in factory:
        name = conf.name
        docs = conf.docs

        arguments = dict()
        arguments['help'] = docs
        arguments['nargs'] = 0

        if len(conf.required_args) or len(conf.optional_args):
            arguments['nargs'] = '+' if len(conf.required_args) else '*'
            optionals = list(map(lambda x: '[%s]' % x, conf.optional_args))
            arguments['metavar'] = tuple(conf.required_args + optionals)

        arguments['action'] = action_with_arguments(action,
                                                    conf.required_args,
                                                    conf.optional_args)

        parser.add_argument('--%s' % (name,), **arguments)


class _ActionWithArguments:
    """
    Placeholder class to recognize an argument is a NormalizerAction in argparse
    """


def action_with_arguments(action, required_args, optional_args):
    """
    Custom argparse action to support a variable amount of arguments
    :param action: name of the action
    :param list required_args: required arguments
    :param list optional_args: optional arguments
    :rtype: ActionWithArguments
    """

    minlen = len(required_args)
    maxlen = minlen + len(optional_args)

    class ActionWithArguments(argparse.Action, _ActionWithArguments):
        def __call__(self, parser, args, values, option_string=None):
            if len(values) < minlen or len(values) > maxlen:
                if minlen == maxlen:
                    lentxt = str(minlen)
                else:
                    lentxt = 'between %d and % d' % (minlen, maxlen)
                raise ArgumentError(self, 'requires %s arguments (got %d)' % (lentxt, len(values)))

            if not hasattr(args, action):
                setattr(args, action, [])

            getattr(args, action).append([self.dest] + values)

    return ActionWithArguments


# TODO: further augment formatter to give cleaner output
class HelpFormatter(argparse.HelpFormatter):
    """
    Custom formatter for argparse that allows us to properly display _ActionWithArguments and docblock documentation,
    as well as allowing newlines inside the description.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._whitespace_matcher = re.compile(r'[\t ]+', re.ASCII)

    def _format_args(self, action, default_metavar):
        if isinstance(action, _ActionWithArguments):
            return ' '.join(action.metavar)

        return super()._format_args(action, default_metavar)

    def _split_lines(self, text, width):
        def wrap(txt):
            if txt == '':
                return ['']
            return textwrap.wrap(txt, width=width)

        text = text.splitlines()
        text = list(itertools.chain.from_iterable(map(wrap, text)))
        return text

    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(' ', text).strip()
        wrapper = partial(textwrap.fill,
                          width=width,
                          initial_indent=indent,
                          subsequent_indent=indent)
        return '\n'.join(map(wrapper, text.split('\n')))


def args_complete(parser):  # pragma: no cover
    try:
        # support argument completion if package is installed
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass


def create_parser(*args, **kwargs):
    defaults = {
        'add_help': False,
        'formatter_class': HelpFormatter,
        'allow_abbrev': False,
    }
    kwargs = {k: kwargs.get(k, v) for k, v in defaults.items()}
    parser = argparse.ArgumentParser(*args, **kwargs)
    parser._optionals.title = 'named arguments'
    return parser


@contextmanager
def main_parser_context():
    try:
        # import done here to avoid circular references
        import benchmarkstt.benchmark.cli as benchmark_cli
        name = 'benchmarkstt'
        desc = 'BenchmarkSTT\'s main command line tool that is used for benchmarking speech-to-text, ' \
               'for additional tools, see ``benchmarkstt-tools --help``.'

        parser = create_parser(prog=name, description=desc)
        benchmark_cli.argparser(parser)

        parser.add_argument('--version', action='store_true',
                            help='Output %s version number' % (name,))

        args_common(parser)
        args_help(parser)
        yield parser
    finally:
        pass


def determine_log_level():
    # Set log-level manually before parse_args(), so that also factory logs, etc. get output
    log_level = 'WARNING'
    if '--log-level' in sys.argv:
        idx = sys.argv.index('--log-level') + 1
        if idx < len(sys.argv):
            if sys.argv[idx].upper() in logging._nameToLevel:
                log_level = sys.argv[idx].upper()

    logging.basicConfig(level=log_level)
    logging.getLogger().setLevel(log_level)


def main():
    determine_log_level()
    # import done here to avoid circular dependencies
    import benchmarkstt.benchmark.cli as benchmark_cli
    with main_parser_context() as parser:
        args_complete(parser)

        if '--version' in sys.argv:
            print("benchmarkstt: %s" % (__meta__.__version__,))
            logging.getLogger().info('python version: %s', sys.version)
            parser.exit(0)

        args = parser.parse_args()
        benchmark_cli.main(parser, args)

    exit(0)


def main_parser():
    with main_parser_context() as parser:
        return parser


def tools_parser():
    name = 'benchmarkstt-tools'
    desc = 'Some additional helpful tools'
    parser = create_parser(prog=name, description=desc)

    subparsers = parser.add_subparsers(dest='subcommand')

    for module, cli in Modules('cli'):
        kwargs = dict()
        if hasattr(cli, 'Formatter'):
            kwargs['formatter_class'] = cli.Formatter
        else:
            kwargs['formatter_class'] = HelpFormatter

        docs = cli.__doc__ if cli.__doc__ is not None else ('TODO: add description to benchmarkstt.%s.cli' % (module,))
        kwargs['description'] = textwrap.dedent(docs)
        subparser = subparsers.add_parser(module, add_help=False, allow_abbrev=False, **kwargs)

        cli.argparser(subparser)
        args_common(subparser)
        args_help(subparser)

    args_help(parser)
    return parser


def tools():
    determine_log_level()
    parser = tools_parser()
    args_complete(parser)

    args = parser.parse_args()

    if not args.subcommand:
        parser.error("expects at least 1 argument")

    Modules('cli')[args.subcommand].main(parser, args)
    exit(0)


if __name__ == '__main__':  # pragma: nocover
    main()



from benchmarkstt.input.core import PlainText
import benchmarkstt.metrics as metrics

factory = metrics.factory


def callback(cls, ref: str, hyp: str, *args, **kwargs):
    """
    :param ref: Reference text
    :param hyp: Hypothesis text
    """
    ref = PlainText(ref)
    hyp = PlainText(hyp)
    return cls(*args, **kwargs).compare(ref, hyp)



"""
Do a complete flow of input -> normalization -> segmentation -> metrics
"""

import benchmarkstt.metrics.cli as metrics_cli
from benchmarkstt.normalization.cli import args_logs, args_normalizers, get_normalizer_from_args
import argparse

# hidden: don't add this as a subcommand to benchmarkstt-tools
hidden = True


def argparser(parser: argparse.ArgumentParser):
    metrics_cli.argparser(parser)
    args_normalizers(parser)
    args_logs(parser)
    return parser


def main(parser, args):
    normalizer = get_normalizer_from_args(args)
    metrics_cli.main(parser, args, normalizer)



import os
import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 5):
    sys.exit('Sorry, Python < 3.5 is not supported')


def from_file(filename):
    with open(filename) as f:
        result = f.read()
    return result.strip()


dirname = os.path.dirname(__file__)
__version__ = from_file('VERSION')
__author__ = 'EBU'
__name__ = 'benchmarkstt'

# Auto save to __meta__
meta_location = os.path.join(dirname, 'src', __name__, '__meta__.py')
with open(meta_location, 'w') as f:
    f.write('''# Automatically created. DO NOT EDIT
__version__ = %s
__author__ = %s
''' % (repr(__version__), repr(__author__)))

long_description = from_file('README.rst')

setup(
    name=__name__,
    url='https://github.com/ebu/benchmarkstt/',
    license='MIT',
    version=__version__,
    author=__author__,
    author_email='ai-stt@list.ebu.ch',
    maintainer=__author__,
    maintainer_email='ai-stt@list.ebu.ch',
    description='A library for benchmarking AI/ML applications.',
    long_description=long_description,
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.5',
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={'benchmarkstt': ['api/templates/*.html']},
    include_package_data=True,
    install_requires=[
        'MarkupSafe>=1.0',
        'Unidecode>=1.1.0',
        'langcodes>=1.4.1',
        'Flask>=1.0.2',
        'jsonrpcserver>=4.0.1',
        'gunicorn>=19.9.0',
        'docutils>=0.14',
    ],
    extras_require={
        'test': [
            "pytest==4.2.0",
            "pycodestyle==2.5.0",
            "pytest-cov==2.5.1"
        ]
    },
    platforms='any',
    entry_points={
        'console_scripts': [
            "%s=%s.cli:main" % (__name__, __name__),
            "%s-tools=%s.cli:tools" % (__name__, __name__)
        ],
    }
)



import logging
from benchmarkstt import make_printable
import difflib
from markupsafe import escape
from benchmarkstt.schema import Schema
from io import StringIO
from collections import OrderedDict

logger = logging.getLogger(__name__)


class Dialect:
    preprocessor = None
    delete_format = '%s'
    insert_format = '%s'
    equal_format = '%s'
    replace_format = None

    def __init__(self):
        self._stream = StringIO()

    @property
    def stream(self):
        return self._stream

    def __enter__(self):
        self._stream = StringIO()
        return self._stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def output(self):
        return self._stream.getvalue()


class CLIDiffDialect(Dialect):
    def __init__(self, show_color_key=None):
        self.show_color_key = bool(show_color_key) if show_color_key is not None else True

        self.color_key = 'Color key: Unchanged %s %s\n\n' % (
            self.delete_format % 'Reference',
            self.insert_format % 'Hypothesis')

        super().__init__()

    @staticmethod
    def preprocessor(txt):
        return make_printable(txt)

    def __enter__(self):
        super().__enter__()
        self._stream = StringIO()
        if self.show_color_key:
            self._stream.write(self.color_key)
        return self

    delete_format = '\033[31m%s\033[0m'
    insert_format = '\033[32m%s\033[0m'


class UTF8Dialect(Dialect):
    @staticmethod
    def preprocessor(txt):
        return make_printable(txt)

    def delete_format(self, txt):
        self._stream.writelines(c + '\u0338' for c in txt)

    def insert_format(self, txt):
        self._stream.writelines(c + '\u0359' for c in txt)


class HTMLDiffDialect(Dialect):
    @staticmethod
    def preprocessor(txt):
        return escape(txt)

    delete_format = '<span class="delete">%s</span>'
    insert_format = '<span class="insert">%s</span>'


class RestructuredTextDialect(CLIDiffDialect):
    @staticmethod
    def preprocessor(txt):
        return CLIDiffDialect.preprocessor(txt).replace('·', '\u200B·\u200B').replace('`', r'\`')

    delete_format = '\\ :diffdelete:`%s`\\ '
    insert_format = '\\ :diffinsert:`%s`\\ '


class ListDialect(Dialect):
    @staticmethod
    def preprocessor(txt):
        return txt

    def delete_format(self, txt):
        return self._format('delete', txt)

    def insert_format(self, txt):
        return self._format('insert', txt)

    def equal_format(self, txt):
        return self._format('equal', txt)

    def replace_format(self, a, b):
        return self._format('replace', a, b)

    def _format(self, kind, txt, txt2=None):
        txt = txt.split()
        if txt2 is None:
            txt2 = txt
        else:
            txt2 = txt2.split()

        oor = len(txt2)
        for idx, word in enumerate(txt):
            ref = word if kind != 'insert' else None
            if idx >= oor:
                kind = 'delete'
            hyp = txt2[idx] if kind != 'delete' else None
            result = OrderedDict((('type', kind), ('reference', ref), ('hypothesis', hyp)))
            self._output.append(result)

        if idx < oor:
            for word_ in txt2[idx+1:]:
                result = OrderedDict((('type', 'insert'), ('reference', None), ('hypothesis', word_)))
                self._output.append(result)

    def __enter__(self):
        self._output = []
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def output(self):
        return self._output


class JSONDiffDialect(ListDialect):
    def __init__(self):
        self._line = None

    def __enter__(self):
        super().__enter__()
        self._stream = StringIO()
        self._line = 0
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        self._line = None
        self._stream.write(Schema.dumps(super().output()))

    def output(self):
        return self._stream.getvalue()


class DiffFormatter:
    diff_dialects = {
        "cli": CLIDiffDialect,
        "html": HTMLDiffDialect,
        "text": UTF8Dialect,
        "json": JSONDiffDialect,
        "list": ListDialect,
        "rst": RestructuredTextDialect,
    }

    def __init__(self, dialect=None, *args, **kwargs):
        if dialect is None:
            dialect = 'text'

        if not self.has_dialect(dialect):
            raise ValueError("Unknown diff dialect", dialect)
        self._dialect = self.diff_dialects[dialect](*args, **kwargs)

    def diff(self, a, b, opcodes=None, preprocessor=None):
        formats = dict(insert=None, delete=None, equal=None, replace=None)

        dialect = self._dialect
        with dialect as stream:
            def format_string(formatting):
                def _(*args):
                    dialect.stream.write(formatting % args)

                return _

            # dialect = self._dialect
            for f in formats.keys():
                formatter = getattr(dialect, f + '_format')
                if type(formatter) is str:
                    formats[f] = format_string(formatter)
                else:
                    formats[f] = formatter

            if formats['replace'] is None:
                def _(deleted, inserted):
                    formats['delete'](deleted)
                    formats['insert'](inserted)
                formats['replace'] = _

            if preprocessor is not None:
                def _pre(txt):
                    return dialect.preprocessor(preprocessor(txt))
            else:
                _pre = dialect.preprocessor

            if opcodes is None:
                opcodes = difflib.SequenceMatcher(None, a, b).get_opcodes()

            for tag, alo, ahi, blo, bhi in opcodes:
                a_ = _pre(a[alo:ahi])

                if tag in ('equal', 'delete'):
                    formats[tag](_pre(a[alo:ahi]))
                else:
                    b_ = _pre(b[blo:bhi])
                    if tag == 'insert':
                        formats[tag](b_)
                    else:
                        formats[tag](a_, b_)
        return dialect.output()

    @classmethod
    def has_dialect(cls, dialect):
        return dialect in cls.diff_dialects


def format_diff(a, b, opcodes=None, dialect=None, preprocessor=None):
    formatter = DiffFormatter(dialect)
    return formatter.diff(a, b, opcodes, preprocessor)



"""
Core segmenters, each segmenter must be Iterable returning a Item
"""

import re
from benchmarkstt.schema import Item
from benchmarkstt.segmentation import Base


class Simple(Base):
    """
    Simplest case, split into words by white space
    """

    def __init__(self, text: str, pattern=r'[\n\t\s]+', normalizer=None):
        self._text = text
        self._re = re.compile('(%s)' % (pattern,))
        self._normalizer = normalizer
        if self._normalizer is not None:
            self._text = self._normalizer.normalize(text)

    def __iter__(self):
        start_match = self._re.match(self._text)
        iterable = self._re.split(self._text)
        if iterable[0] == '':
            iterable.pop(0)

        pos = 0
        length = len(iterable)

        # special case, starts with word break, add it to first word
        if start_match is not None:
            matches = iterable[0:3]
            pos = 3
            yield Item({"item": matches[1], "type": "word", "@raw": ''.join(matches)})

        while pos < length:
            raw = ''.join(iterable[pos:pos+2])
            if raw != '':
                yield Item({"item": iterable[pos], "type": "word", "@raw": raw})
            pos += 2



from benchmarkstt.normalization.logger import LogCapturer
import json
import benchmarkstt.csv as csv
import benchmarkstt.normalization as normalization

factory = normalization.factory


def callback(cls, text: str, return_logs: bool = None, *args, **kwargs):
    """
    :param text: The text to normalize
    :param bool return_logs: Return normalization logs
    """
    try:
        instance = cls(*args, **kwargs)
        if not return_logs:
            return dict(text=instance.normalize(text))

        with LogCapturer(dialect='html', diff_formatter_dialect='dict') as logcap:
            result = dict(text=instance.normalize(text))
            result['logs'] = logcap.logs
            return result
    except csv.CSVParserError as e:
        message = 'on line %d, character %d' % (e.line, e.char)
        message = '\n'.join([e.__doc__, e.message, message])
        data = {
            "message": message,
            "line": e.line,
            "char": e.char,
            "index": e.index,
            "field": "config"
        }
        raise AssertionError(json.dumps(data))



"""
Subpackage responsible for dealing with output formats
"""

from benchmarkstt.factory import Factory


class Base:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def result(self, title, result):
        raise NotImplementedError()


factory = Factory(Base)



import inspect
import logging
from importlib import import_module
from benchmarkstt.docblock import format_docs
from collections import namedtuple
from typing import Dict
from benchmarkstt.registry import Registry
from benchmarkstt.modules import load_object

logger = logging.getLogger(__name__)


class ClassConfig(namedtuple('ClassConfigTuple', ['name', 'cls', 'docs', 'optional_args', 'required_args'])):
    @property
    def docs(self):
        if self.cls.__doc__ is None:
            docs = ''
            logger.warning("No docstring for '%s'", self.name)
        else:
            docs = self.cls.__doc__
        return format_docs(docs)


class Factory(Registry):
    """
    Factory class with auto-loading of namespaces according to a base class.
    """

    def __init__(self, base_class, namespaces=None):
        super().__init__()
        self.base_class = base_class
        if namespaces is None:
            self.namespaces = [base_class.__module__ + '.core']
        else:
            self.namespaces = namespaces

        for namespace in self.namespaces:
            self.register_namespace(namespace)

    def __contains__(self, item):
        return super().__contains__(self.normalize_class_name(item))

    def __getitem__(self, item):
        """
        Loads the proper class based on a name

        :param name: Case-insensitive name of the class
        :return: The class
        :rtype: class
        """
        name = self.normalize_class_name(item)

        if name not in self._registry:
            raise ImportError("Could not find class '%s', available: %s" % (name, ', '.join(self._registry.keys())))

        return super().__getitem__(name)

    def __delitem__(self, key):
        if type(key) is not str:
            key = self.normalize_class_name(key.__name__)
        super().__delitem__(key)

    def create(self, alias, *args, **kwargs):
        return self[alias](*args, **kwargs)

    @staticmethod
    def normalize_class_name(clsname):
        """
        Normalizes the class name for automatic lookup of a class, by default
        this means lowercasing the class name, but may be overrided by a child
        class.

        :param clsname: The class name
        :return: The normalized class name
        :rtype: str
        """
        return clsname.lower()

    def is_valid(self, tocheck):
        """
        Checks that tocheck is a valid class extending base_class

        :param tocheck: The class to check
        :rtype: bool
        """

        if tocheck is self.base_class:
            return False
        if not inspect.isclass(tocheck):
            return False
        if not issubclass(tocheck, self.base_class):
            return False

        return True

    def register_namespace(self, namespace):
        """
        Registers all valid classes from a given namespace

        :param str|module namespace:
        """

        module = '.'.join(filter(len, namespace.split('.')))
        if module == '':
            module = globals()
        else:
            module = import_module(module)

        for clsname in dir(module):
            cls = getattr(module, clsname)
            if not self.is_valid(cls):
                continue
            self.register(cls, clsname)

    def register_classname(self, name, alias=None):
        if alias is None:
            alias = name
        self.register(load_object(name), alias)

    def register(self, cls, alias=None):
        """
        Register an alias for a class

        :param self.base_class cls:
        :param str|None alias: The alias to use when trying to get the class back,
                               by default will use normalized class name.
        :return: None
        """
        if not self.is_valid(cls):
            raise ValueError('Invalid class (must inherit from Base class)"')

        if alias is None:
            alias = cls.__name__

        alias = self.normalize_class_name(alias)
        super().register(alias, cls)

    def __iter__(self):
        """
        Get available classes with a proper ClassConfig

        :return: A dictionary of registered classes
        :rtype: Dict[str, ClassConfig]
        """

        for clsname, cls in self._registry.items():
            argspec = inspect.getfullargspec(cls.__init__)
            args = list(argspec.args)[1:]
            defaults = []
            if argspec.defaults:
                defaults = list(argspec.defaults)

            defaults_idx = len(args) - len(defaults)
            required_args = args[0:defaults_idx]
            optional_args = args[defaults_idx:]

            yield ClassConfig(name=clsname, cls=cls,
                              docs=None,
                              optional_args=optional_args,
                              required_args=required_args)
